from django.conf import settings


class Settings:
    """A helper class to access settings in a more convenient way."""

    @classmethod
    def _get(cls, item):
        result = getattr(settings, "DJANGO_FORBID", {})
        for attr in item.split("."):
            result = result[attr]
        return result

    @classmethod
    def has(cls, item):
        try:
            cls._get(item)
            return True
        except KeyError:
            return False

    @classmethod
    def get(cls, item, default=None):
        try:
            return cls._get(item)
        except KeyError:
            return default



from django.conf import settings
from django.contrib.gis.geoip2 import GeoIP2
from geoip2.errors import AddressNotFoundError

from .config import Settings


class Rule:
    # Key in the geoip2 city object.
    # Subclasses should override this.
    key = None

    def __init__(self, code):
        # Two-letter ISO 3166-1 alpha-2 code.
        self.code = code

    def __call__(self, city):
        """Checks if the code is satisfied."""
        return self.code == city.get(self.key)


class CountryRule(Rule):
    key = "country_code"


class ContinentRule(Rule):
    key = "continent_code"


class Access:
    countries = "COUNTRIES"
    territories = "TERRITORIES"

    # Hold the instance of GeoIP2.
    geoip = GeoIP2()

    def __init__(self):
        self.rules = []

        if Settings.has(self.countries):
            for country in Settings.get(self.countries):
                self.rules.append(CountryRule(country.upper()))

        if Settings.has(self.territories):
            for territory in Settings.get(self.territories):
                self.rules.append(ContinentRule(territory.upper()))

    def accessible(self, city):
        """Checks if the IP address is in the white zone."""
        return any(map(lambda rule: rule(city), self.rules))

    def grants(self, city):
        """Checks if the IP address is permitted."""
        raise NotImplementedError


class PermitAccess(Access):
    def grants(self, city):
        """Checks if the IP address is permitted."""
        return not self.rules or self.accessible(city)


class ForbidAccess(Access):
    def grants(self, city):
        """Checks if the IP address is forbidden."""
        return not self.rules or not self.accessible(city)


class Factory:
    """Creates an instance of the Access class."""

    FORBID = ForbidAccess
    PERMIT = PermitAccess

    @classmethod
    def create_access(cls, action):
        return getattr(cls, action)()


def grants_access(request, ip_address):
    """Checks if the IP address is in the white zone."""
    try:
        city = Access.geoip.city(ip_address)

        # Saves the timezone in the session for
        # comparing it with the timezone in the
        # POST request sent from user's browser
        # to detect if the user is using VPN.
        timezone = city.get("time_zone")
        request.session["tz"] = timezone

        # Creates an instance of the Access class
        # and checks if the IP address is granted.
        action = Settings.get("OPTIONS.ACTION", "FORBID")
        return Factory.create_access(action).grants(city)
    except (AddressNotFoundError, Exception):
        # This happens when the IP address is not
        # in  the  GeoIP2 database. Usually, this
        # happens when the IP address is a local.
        return not any([
            Settings.has(Access.countries),
            Settings.has(Access.territories),
        ]) or getattr(settings, "DEBUG", False)



from datetime import datetime

from django.http import HttpResponseForbidden
from django.shortcuts import redirect
from django.utils.timezone import utc

from .access import grants_access
from .config import Settings
from .detect import detect_vpn


class ForbidMiddleware:
    """Middleware to forbid access to the site."""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        address = request.META.get("REMOTE_ADDR")
        address = request.META.get("HTTP_X_FORWARDED_FOR", address)

        # Checks if the PERIOD attr is set and the user has been granted access.
        if Settings.has("OPTIONS.PERIOD") and request.session.has_key("ACCESS"):
            acss = datetime.utcnow().replace(tzinfo=utc).timestamp()

            # Checks if access is not timed out yet.
            if acss - request.session.get("ACCESS") < Settings.get("OPTIONS.PERIOD"):
                return detect_vpn(self.get_response, request)

        # Checks if access is granted when timeout is reached.
        if grants_access(request, address.split(",")[0].strip()):
            acss = datetime.utcnow().replace(tzinfo=utc)
            request.session["ACCESS"] = acss.timestamp()
            return detect_vpn(self.get_response, request)

        # Redirects to the FORBIDDEN_LOC URL if set.
        if Settings.has("OPTIONS.URL.FORBIDDEN_LOC"):
            return redirect(Settings.get("OPTIONS.URL.FORBIDDEN_LOC"))

        return HttpResponseForbidden()



import json
import re

from django.http import HttpResponse
from django.http import HttpResponseForbidden
from django.shortcuts import redirect
from django.shortcuts import render

from .config import Settings


def detect_vpn(get_response, request):
    response_attributes = ("content", "charset", "status", "reason")

    def erase_response_attributes():
        for attr in response_attributes:
            request.session.pop(attr)

    if any([
        # The session key is checked to avoid
        # redirect loops in development mode.
        not request.session.has_key("tz"),
        # Checks if VPN is False or not set.
        not Settings.get("OPTIONS.VPN", False),
        # Checks if the request is an AJAX request.
        not re.search(
            r"\w+\/(?:html|xhtml\+xml|xml)",
            request.META.get("HTTP_ACCEPT"),
        ),
    ]):
        return get_response(request)

    if all(map(request.session.has_key, ("tz", *response_attributes))):
        # Handles if the user's timezone differs from the
        # one determined by GeoIP API. If so, VPN is used.
        if request.POST.get("timezone", "N/A") != request.session.get("tz"):
            erase_response_attributes()
            # Redirects to the FORBIDDEN_VPN URL if set.
            if Settings.has("OPTIONS.URL.FORBIDDEN_VPN"):
                return redirect(Settings.get("OPTIONS.URL.FORBIDDEN_VPN"))
            return HttpResponseForbidden()

        # Restores the response from the session.
        response = HttpResponse(**{attr: request.session.get(attr) for attr in response_attributes})
        if hasattr(response, "headers"):
            response.headers = json.loads(request.session.get("headers"))
        erase_response_attributes()
        return response

    # Gets the response and saves attributes in the session to restore it later.
    response = get_response(request)
    if hasattr(response, "headers"):
        # In older versions of Django, HttpResponse does not have headers.
        request.session["headers"] = json.dumps(dict(response.headers))
    request.session["content"] = response.content.decode(response.charset)
    request.session["charset"] = response.charset
    request.session["status"] = response.status_code
    request.session["reason"] = response.reason_phrase

    return render(request, "timezone.html", status=302)
